#include <stdint.h>

//#include "ecoprimer.h"
#include "PrimerSets.h"

int32_t counttaxon(int32_t taxid);

float get_set_coverage (pairset *p_set, SetParams *pparams, int32_t pidx_toexclude)
{
	int32_t i, j;
	float cov;
	pprimer_t prmr1;
	pprimer_t prmr2;
	int32_t s_intaxa = 0;
	
	counttaxon(-1);
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
	{
		if (p_set->set_pairs[i] == -1 || pidx_toexclude == i) continue;
		
		prmr1 = pparams->sortedpairs[p_set->set_pairs[i]]->p1;
		prmr2 = pparams->sortedpairs[p_set->set_pairs[i]]->p2;
		for (j = 0; j < pparams->options->dbsize; j++)
			if ((prmr1->directCount[j] > 0 || prmr1->reverseCount[j] > 0)
					&& (prmr2->directCount[j] > 0 || prmr2->reverseCount[j] > 0)
					&& pparams->seqdb[j]->isexample && pparams->seqdb[j]->ranktaxonid > 0)
			{
				s_intaxa = counttaxon(pparams->seqdb[j]->ranktaxonid);
			}
	}
	//fprintf(stderr, "%d/%d\n", s_intaxa, pparams->options->intaxa);
	cov = s_intaxa*1.0/(pparams->options->intaxa*1.0);

	return cov;
}

void set_cov_spc (pairset *p_set, SetParams *pparams)
{
	int32_t i;
	int32_t ssp = 0;
	
	for (i = 0; i < pparams->options->dbsize; i++)
		if (p_set->set_wellIdentifiedTaxa[i] == 1) ssp++;
	
	//set specificity
	p_set->set_specificity = ssp*1.0/(pparams->options->insamples*1.0);
	
	//set coverage
	p_set->set_coverage = get_set_coverage (p_set, pparams, -1);
}

int ok_to_add (int id, pairset *pair_set)
{
	int i;
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
		if (pair_set->set_pairs[i] == id) return 0;
	
	return 1;
}

//1. Add in set the first pair having highiest specificity
//2. For the sequences not WI by primers in set, calculate count for each pair equal to number of seqs WI by it
//3. Take the pair with highiest such count and see its links with primers already in set, 
//4. If no/insufficient links, take next pair
//5. repeate 3,4 untill pair gets added in set
//6. Repeate 2 to 5 until set complete

pairset build_primers_set (ppair_t* sortedpairs, int32_t sorted_count, pecodnadb_t seqdb,
		poptions_t options)
{
	SetParams params;
	pairset pair_set;
	int32_t i;
	int32_t pset_idx;
	int32_t prb_idx;
	int *pair_wi_count_sorted_ids;
	
	params.sortedpairs = sortedpairs;
	params.sorted_count = sorted_count;
	params.seqdb = seqdb;
	params.options = options;

	memset (&pair_set, 0, sizeof(pairset));
	pair_set.set_wellIdentifiedTaxa = (int *) ECOMALLOC(options->dbsize*sizeof (int),
            "Could not allocate memory for pair set");
	
	pair_wi_count_sorted_ids = (int *) ECOMALLOC(sorted_count*sizeof (int),
            "Could not allocate memory for pair_wi_count_sorted");
	
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
		pair_set.set_pairs[i] = -1;
	
	pset_idx = 0;
	prb_idx = 0;
	//add first pair by default, this should be the one having highiest specificty
	add_pair_in_set (&pair_set, pset_idx, prb_idx, &params);
	pset_idx++;
	
	while (pset_idx < PRIMERS_IN_SET_COUNT)
	{
		//get a sorted list of pair ids with the pair well identifying most of the remaining seqs at top
		get_next_pair_options (pair_wi_count_sorted_ids, &pair_set, &params);
		
		if (pair_wi_count_sorted_ids[0] == 0)
		{
			fprintf (stderr, "No further pair found, total so far %d\n", pset_idx);
			break;
		}
		else
			fprintf (stderr, "Some counts found\n");
		
		
		for (prb_idx = 0; prb_idx < sorted_count; prb_idx++)
		{
			if (pair_wi_count_sorted_ids[prb_idx])
			if (ok_to_add (pair_wi_count_sorted_ids[prb_idx], &pair_set))
			{
				//fprintf (stderr, "Oktoadd\n");
				float ldist = get_links_distribution (pair_wi_count_sorted_ids[prb_idx], &pair_set, &params);
				fprintf (stderr, "ldist:%f\n", ldist);
				if (ldist >= 0.7)
				{
					add_pair_in_set (&pair_set, pset_idx, pair_wi_count_sorted_ids[prb_idx], &params);
					pset_idx++;
					break;
				}
				else
					fprintf (stderr, "ldist:%f\n", ldist);
			}
		}
		
		if (prb_idx == sorted_count)
		{
			fprintf (stderr, "No further pair found, total so far %d\n", pset_idx);
			break;
		}
		
		if (pair_set.set_specificity == 1.0)
		{
			fprintf (stderr, "Set Complete with total primers: %d\n", pset_idx);
			break;
		}

	}
	get_set_mean_cov_stats (&pair_set, &params);
	
	return pair_set;
}

float get_links_distribution (int prb_idx, pairset *pair_set, SetParams *pparams)
{
	int i, j;
	int *pair_link_count;
	int *pwi;
	int *pswi;
	float pcnt = 0.0;
	float pscnt = 0.0;
	
	pair_link_count = (int *) ECOMALLOC(PRIMERS_IN_SET_COUNT*sizeof (int),
	            "Could not allocate memory for pair_link_count");
	
	pwi = pparams->sortedpairs[prb_idx]->wellIdentifiedSeqs;
	//fprintf(stderr, "a,");
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
	{
		pair_link_count[i] = 0;
		if (pair_set->set_pairs[i] != -1)
		{
			//fprintf(stderr, "b,");
			pswi = pparams->sortedpairs[pair_set->set_pairs[i]]->wellIdentifiedSeqs;
			for (j = 0; j < pparams->options->dbsize; j++)
				if (pwi[j] == 1 && pwi[j] == pswi[j])
					pair_link_count[i] += 1;
		}
	}
	//fprintf(stderr, "c,");
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
	{
		if (pair_set->set_pairs[i] != -1)
			pscnt++;
		if (pair_link_count[i] > 0)
			pcnt++;
	}
	ECOFREE (pair_link_count, "Could not free memory for pair_link_count");
	//fprintf(stderr, "d,");
	return (pcnt/pscnt);
}


void get_next_pair_options (int *pair_wi_count_sorted_ids, pairset *pair_set, SetParams *pparams)
{
	int *pair_count;
	int32_t i, j;
	int max;
	int tmp;
	
	pair_count = (int *) ECOMALLOC(pparams->sorted_count*sizeof (int),
            "Could not allocate memory for pair_count");
	
	memset (pair_wi_count_sorted_ids, 0, pparams->sorted_count*sizeof(int));
	
	for (i = 0; i < pparams->options->dbsize; i++)
	{
		if (pair_set->set_wellIdentifiedTaxa[i] == 1) continue;
		
		for (j = 0; j < pparams->sorted_count; j++)
		{
			if (pparams->sortedpairs[j]->wellIdentifiedSeqs[i] == 1)
				pair_count[j] += 1;
		}
	}
	
	//set pair ids
	for (j = 0; j < pparams->sorted_count; j++)
		pair_wi_count_sorted_ids[j] = j;
	
	//set count of primers already in set to zero (it should already be zero) just for testing
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
		if (pair_set->set_pairs[i] != -1)
			pair_count[pair_set->set_pairs[i]] = 0;
	
	//sort two arrays in descending wrt count
	for (i = 0; i < pparams->sorted_count - 1; i++)
	{
		max = i;
		for (j = i + 1; j < pparams->sorted_count; j++)
			if (pair_count[max] < pair_count[j])
				max = j;
		
		if (max > i)
		{
			tmp = pair_count[i];
			pair_count[i] = pair_count[max];
			pair_count[max] = tmp;
			
			tmp = pair_wi_count_sorted_ids[i];
			pair_wi_count_sorted_ids[i] = pair_wi_count_sorted_ids[max];
			pair_wi_count_sorted_ids[max] = tmp;
		}
	}
	
	for (i = 0; i < pparams->sorted_count - 1; i++)
		if (pair_count[i] == 0)
			pair_wi_count_sorted_ids[i] = 0;	
		//else
		//	fprintf (stderr, "%d:%d, ", i, pair_count[i]);
	
	ECOFREE (pair_count, "Could not free memory for pair_count");
}

void add_pair_in_set (pairset *pair_set, int32_t pset_idx, int32_t prb_idx, SetParams *pparams)
{
	int *pwi;
	int32_t i;
	
	pair_set->set_pairs[pset_idx] = prb_idx;
	pwi = pparams->sortedpairs[prb_idx]->wellIdentifiedSeqs;
	for (i = 0; i < pparams->options->dbsize; i++)
		if (pwi[i] == 1)
			pair_set->set_wellIdentifiedTaxa[i] = 1;
	
	set_cov_spc (pair_set, pparams);
}

void get_set_mean_cov_stats (pairset *pair_set, SetParams *pparams)
{
	int32_t i, j, k;
	int interseq_vals[PRIMERS_IN_SET_COUNT*PRIMERS_IN_SET_COUNT];
	int interseq_cnt = 0;
	double msum;
	int *p1wi;
	int *p2wi;
	
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
	{
		if (pair_set->set_pairs[i] == -1) continue;
		p1wi = pparams->sortedpairs[pair_set->set_pairs[i]]->wellIdentifiedSeqs;
		for (j = i+1; j < PRIMERS_IN_SET_COUNT; j++)
		{
			if (pair_set->set_pairs[j] == -1) continue;
			p2wi = pparams->sortedpairs[pair_set->set_pairs[j]]->wellIdentifiedSeqs;
			interseq_vals[interseq_cnt] = 0;
			
			for (k = 0; k < pparams->options->dbsize; k++)
				if (p1wi[k] == 1 && p2wi[k] == 1)
					interseq_vals[interseq_cnt]++;
			interseq_cnt++;
		}
	}
	
	//calculate mean
	msum = 0;
	pair_set->set_score = 0;
	pair_set->set_lmean = 0;
	pair_set->set_lcov = -1;
	if (interseq_cnt == 0) return;
	
	for (i = 0; i < interseq_cnt; i++)
		msum += interseq_vals[i];
	pair_set->set_lmean = msum/interseq_cnt;
	
	msum = 0;
	for (i = 0; i < interseq_cnt; i++)
		msum += (interseq_vals[i] - pair_set->set_lmean)*(interseq_vals[i] - pair_set->set_lmean);
	pair_set->set_lcov = msum/interseq_cnt;
	
	if (pair_set->set_lcov != 0)
		pair_set->set_score = (pair_set->set_coverage*pair_set->set_specificity*pair_set->set_lmean)/pair_set->set_lcov;
}

void reset_set_props (pairset *pair_set, SetParams *pparams)
{
	int *pwi;
	int i, j;
	
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
	{
		if (pair_set->set_pairs[i] == -1) continue;
		pwi = pparams->sortedpairs[pair_set->set_pairs[i]]->wellIdentifiedSeqs;
		for (j = 0; j < pparams->options->dbsize; j++)
			if (pwi[j] == 1)
				pair_set->set_wellIdentifiedTaxa[j] = 1;
	}
	
	set_cov_spc (pair_set, pparams);
	get_set_mean_cov_stats (pair_set, pparams);
}

void print_set_info (pairset *pair_set, SetParams *pparams)
{
	int i;
	int printed1st = 0;
	
	printf ("%f\t%f\t%f\t%f\t%f\t", pair_set->set_specificity,
			pair_set->set_coverage,pair_set->set_lmean,
			pair_set->set_lcov,pair_set->set_score);
	
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
	{
		if (pair_set->set_pairs[i] == -1) continue;
		
		if (printed1st)
			printf (":%d", pair_set->set_pairs[i]);
		else
			printf ("%d", pair_set->set_pairs[i]);
		printed1st = 1;
	}
	printf ("\n");
}

void some_other_set_possibilities (pairset *pair_set,
		ppair_t * sortedpairs, int32_t sorted_count, pecodnadb_t seqdb, poptions_t options)
{
	SetParams params;
	pairset tmp_pair_set;
	int i, j;
	int *wi;
	
	params.sortedpairs = sortedpairs;
	params.sorted_count = sorted_count;
	params.seqdb = seqdb;
	params.options = options;
	wi = (int *) ECOMALLOC(options->dbsize*sizeof (int),
					            "Could not allocate memory for pair set wi");
	//print stats for first original set
	printf ("\nspecificity\tcoverage\tmean\tcovariance\tscore\tprimers\n");
	print_set_info (pair_set, &params);
	
	for (i = 0; i < PRIMERS_IN_SET_COUNT; i++)
	{
		if (pair_set->set_pairs[i] == -1) continue;
			
		for (j = 0; j < sorted_count; j++)
		{
			if (ok_to_add (j, pair_set))
			{
				tmp_pair_set = *pair_set;
				tmp_pair_set.set_pairs[i] = j;
				memset (wi, 0, options->dbsize*sizeof (int));
				tmp_pair_set.set_wellIdentifiedTaxa = wi;
				reset_set_props (&tmp_pair_set, &params);
				print_set_info (&tmp_pair_set, &params);
			}
		}
	}
	ECOFREE (wi, "Could not free memory for pair set wi");
}

